/// @file        SDLGame.hpp
/// @description Gamestate Class Wrapper for SDL
/// @author      Joshua O'Leary
/// @revision    2
/// @edited      2023-08-14
/// @license     Unlicense - Public Domain

#pragma once

#include "JUI/Types/Event.hpp"
#include <JUI/common.h>
#include <SDL2/SDL.h>

#include <SDL_image.h>

using namespace JUI;

// Class Wrapper for SDL, plus core game state.
class SDLGame {
#pragma region Interface
public:
#pragma region Constructors
    SDLGame();
    virtual ~SDLGame() {}
#pragma endregion
#pragma region Events
    Event<float> RenderTick;
    Event<float> LogicTick;
    Event<SDL_QuitEvent>             Quitting;
    Event<SDL_WindowEvent>           WindowChanged; // Triggers when **any** window event happens.
    Event<SDL_WindowEvent>           FocusGained;
    Event<SDL_WindowEvent>           FocusLost;
    Event<SDL_WindowEvent>           MouseEntered;
    Event<SDL_WindowEvent>           MouseLeft;
    Event<SDL_KeyboardEvent>         KeyPressed;
    Event<SDL_KeyboardEvent>         KeyReleased;
    Event<SDL_MouseButtonEvent>      MousePressed;
    Event<SDL_MouseButtonEvent>      MouseReleased;
    Event<SDL_MouseMotionEvent>      MouseMoved;
    Event<SDL_MouseWheelEvent>       MouseWheelMoved;
    Event<SDL_ControllerButtonEvent> ButtonPressed;
    Event<SDL_ControllerButtonEvent> ButtonReleased;
    Event<SDL_ControllerAxisEvent>   ThumbstickMoved;
    Event<SDL_JoyButtonEvent>        JoyPressed;
    Event<SDL_JoyButtonEvent>        JoyReleased;
    Event<SDL_TouchFingerEvent>      TouchGesture;
    Event<SDL_TouchFingerEvent>      TouchPressed;
    Event<SDL_TouchFingerEvent>      TouchReleased;
    Event<SDL_DisplayEvent>          DisplayChanged;
    Event<SDL_DisplayEvent>          DisplayConnected;
    Event<SDL_DisplayEvent>          DisplayDisconnected;
    Event<SDL_DisplayEvent>          DisplayOrientationChanged;
    Event<SDL_Renderer*>             OnPreRender;
    Event<SDL_Renderer*>             OnPostRender;
    Event<float>                     OnUpdate;
#pragma endregion
#pragma region Member Methods
#pragma region Virtual Methods (Override Me!)
    virtual void Initialize();        // Derive from this class and override this method!!
    virtual void Cleanup();           // Derive from this class and override this method!!
    virtual void Update(double delta); // Derive from this class and override this method!!
    virtual void Draw();              // Derive from this class and override this method!!
#pragma endregion
#pragma region Getters
    float GetLastFrameTime() const { return frameDelta; }
    float GetFramerate()     const { return frames_per_second; }
    bool    GetFocused()     const;
    v2i GetWindowSize()      const;
    bool    GetCapFPSUnfocused() const;
    v2f GetMouseCoordinates() {
        int x = 0;
        int y = 0;
        SDL_GetMouseState(&x, &y);
        v2f mc{x, y};
        return mc;
    };
    std::string GetWindowTitle() const;
    v2f GetMouseDelta() const { return mouse_delta;}

#pragma endregion
#pragma region Setters
    void SetWindowTitle(const std::string& title);
    void SetWindowSize(int x, int y);
    void SetWindowSize(v2f const &v);
    void SetCappedFPSLimit(int cap);
    void SetCapFPSUnfocused(bool noCap);
    // While the mouse is in relative mode:
    // * The cursor is hidden
    // * The mouse position is constrained to the window
    // * SDL will report continuous relative mouse motion
    void SetMouseRelativeMode(bool enabled);
#pragma endregion
    void Run(); // Starts Render & Gametic Threads
    void Gameloop();
    void RunFrame();
    void Render();
#pragma endregion
#pragma endregion
#pragma region Internals
protected:
    float frames_per_second{}; // FPS measurement as-of last frame
    // TODO: Need to average several measurements later
    bool requestQuit;
    float frameDelta;
    float tickDelta;
    int frameCount;
    bool focused;
    bool capFPSUnfocused = true;
    int cappedFPSLimit = 15;
    v2f mouse_delta;
    v2f last_frame_mouse_pos;


    bool capFPS = false;
    bool vsync = false;
    int fpsLimit = 60;

    SDL_Event event{};
    SDL_Renderer *renderer;
    SDL_Window *window;


    void handleEvents();

#pragma endregion
};