/// @file       Bezier.hpp
/// @desc       Cubic Bezier Curve impl. using De Casteljau's Method
/// @author     Joshua O'Leary
/// @edited     26 October 2023
/// @version    1
/// @copyright  (c) Conarium Software 2023
/// @license    Unlicense - Public Domain

#pragma once

#include <cassert>
#include <cmath>
#include <vector>
#include <limits>
#include <algorithm>
#include <array>
#include <SDL2/SDL_render.h>
#include <glm/glm.hpp>
#include <JUI/common.h>
#include <JUI/math.h>

// Bezier Method:
// p = (1-t)^3 * P0 + 3*t*(1-t)^2*P1 + 3*t^2*(1-t)*P2 + t^3*P3;

// For cubics:
// p = (1-t)^2 * P0 + 2*(1-t)*t*P1 + t*t*P2;

// Transcribed from here: explicit form and derivative
// https://en.wikipedia.org/wiki/B%C3%A9zier_curve#Cubic_B%C3%A9zier_curves

namespace JUI {

    template <typename T> inline T LengthSquared(const v2f& v)
    {
        return (v.x * v.x) + (v.y * v.y);
    }

    template <typename T> inline T Length(const v2f& v) {
        return sqrt(LengthSquared<T>(v));
    }

    // TODO: Implement is_glm_vec concept
    // TODO: Full template parameterization
    inline v2f Normalized(const v2f& val)
    {
        return val / Length<float>(val);
    }

    template <typename T>
    inline T Bezier(float t, T p0, T p1, T p2, T p3)
    {
        return Cube(1 - t) * p0 + 3 * Square(1 - t) * t * p1 + 3 * (1 - t) * Square(t) * p2 + Cube(t) * p3;
    }

    template <typename T>
    inline v2f Bezier(float t, const v2f& p0, const v2f& p1, const v2f& p2, const v2f& p3) {
        return {Bezier(t, p0.x, p1.x, p2.x, p3.x), Bezier(t, p0.y, p1.y, p2.y, p3.y)};
    }

    template <typename T>
    inline T BezierDerivative(float t, T p0, T p1, T p2, T p3) {
        return 3 * Square(1 - t) * (p1 - p0) + 6 * (1 - t) * t * t * (p2 - p1) + 3 * Square(t) * (p3 - p2);
    }

    // Tangent
     v2f BezierDerivative(float t, const v2f& p0, const v2f& p1, const v2f& p2, const v2f& p3);

    // Normal
     v2f BezierNormal(float t, const v2f& p0, const v2f& p1, const v2f& p2, const v2f& p3);

    void DrawCurve(SDL_Renderer* renderer, SDL_Color col, const v2f& p0, const v2f& p1, const v2f p2, const v2f p3, int subdivisions);

    void DrawCurve(SDL_Renderer* renderer, SDL_Color col, const curve& c, int subdivisions);

}