/* @file       Event.hpp
 * @desc       Templated Event Signalling Library
 * @auth       Josh O'Leary
 * @revision   1
 * @edited     2023-05-18
 * @copyright  (C) 2023 Conarium Software
 * @license    Unlicense - Public Domain
 */

#pragma once

#include <functional>
#include <atomic>
#include <list>
#include <thread>
#include <memory>
#include <iostream>
#include <utility>

namespace JUI {
    /// Templated Event Signaller Class
    // The purpose of events is to convey changes in "Class-State" That outside users may be interested in.
    //
    // Example Usage:
    // class Entity {
    // public:
    //      Event<int> OnTick;
    //      void Update(double delta) {
    //          OnTick.Invoke(delta);
    //      }
    // }
    // Example consumption:
    // Entity e = Entity();
    // auto conn = e.OnTick += (float)(delta) { ... };
    // conn.Disconnect();
    //
    template<typename ... Args>
    class Event;


    // Event Delegate
    template<typename ... Args>
    class EventConnection;


    template<typename ... Args>
    class EventConnection {
    private:
        using delegate = std::function<void(Args...)>;

    public:
#pragma region Interface
        EventConnection(Event<Args...> *creator, delegate cb) : owner(creator), callback(std::move(cb)) {}

        bool Disconnect(); // Breaks the event connection, but does not destroy the instance.
        void Invoke(Args... e);
#pragma endregion
#pragma region Implementation
    private:
        Event<Args...> *owner;
        delegate callback;
        bool active = true;
#pragma endregion
    };

    template<typename ... Args>
    class Event {
    private:
        using delegate = std::function<void(Args...)>;
        using connection = EventConnection<Args...>;
        using event_ptr = std::shared_ptr<connection>;

    public:
#pragma region Interface
        void Await(Args &... arg); // Pass in Args... as reference to be filled???
        void Invoke(Args... args);
        void operator()(Args... args) { Invoke(args...); }
        connection Connect(delegate callback);
        void Disconnect(connection &conn);
        connection operator+=(delegate callback) { return Connect(callback); }

#pragma endregion
    private:
        std::vector <event_ptr> listeners;
        uint64_t listenerCounter = 0;
    };

    template<typename... Args>
    void EventConnection<Args...>::Invoke(Args... e) { callback(e...); }

    template<typename ... Args>
    bool EventConnection<Args...>::Disconnect() {
        if (active) {
            owner->Disconnect(this);
            active = false;
            return true;
        }
        return false;
    }

    template<typename ... Args>
    void Event<Args...>::Invoke(Args... args) {
        for (event_ptr &connection_ptr: this->listeners) {
            connection_ptr->Invoke(args...);
        }
    }

    template<typename ... Args>
    typename Event<Args...>::connection Event<Args...>::Connect(delegate callback) {
        event_ptr return_val(new connection(this, callback));
        this->listeners.push_back(return_val);
        return *return_val;
    }

    template<typename ... Args>
    void Event<Args...>::Disconnect(connection &conn) {
        listeners.erase(std::remove(
                listeners.begin(), listeners.end(), 99), listeners.end());
    }


}