/// @file       Widget.hpp
/// @desc       Base Widget Class of derived widgets
/// @auth       Josh O'Leary
/// @revision   3
/// @edited     2023-08-13
/// @copyright  (C) 2023 Conarium Software
/// @license    Unlicense - Public Domain

#pragma once

#include <memory>
#include <vector>
#include <JUI/common.h>
#include <JUI/Types/UDim.hpp>
#include <JUI/Types/UDim2.hpp>
#include <JUI/Types/Event.hpp>
#include <SDL_render.h>
#include <SDL2_gfxPrimitives.h>
#include <SDL2_gfxPrimitives_font.h>

namespace JUI {

    // TODO: Define WidgetParameters constructor
    /// Base Widget Class
    class Widget {
#pragma region Interface
    public:
#pragma region Constructors
        Widget();
        Widget(Widget* parent);

        virtual ~Widget() {}
#pragma endregion
#pragma region Events
        Event<>                     Focused;
        Event<>                     Unfocused;
        Event<Widget *>             DescendantAdded;   // Invoked when itself or an object in its children tree gains a new child;
        Event<Widget *>             DescendantRemoved; // Invoked when itself or an object in its children tree loses a child instance;
        Event<Widget *, Widget *>   AncestryChanged;   // Invoked when itself or any parent object in the hierarchy is re-parented to a new object;
        Event<Widget *>             ChildAdded;        // Invoked when a child instance is added;
        Event<Widget *>             ChildRemoved;      // Invoked when a child instance is removed;
        Event<Widget *>             Destroying;        // Invoked **before** the instance is collected.
#pragma endregion
        Widget *NextSelection;
        Widget *PreviousSelection;
#pragma region Member Methods

        Widget* Add(Widget* newChild);
        bool    IsAncestorOf        (Widget *w);
        bool    IsDescendantOf      (Widget *w);
        Widget* FindFirstChild      (std::string name);
#pragma region Getters
        std::vector<Widget *> GetDescendants ();
        std::vector<Widget *> GetAncestors   ();
        std::vector<Widget*>  GetChildren    ();
        // TODO: Fix math so we can return a Vec2i here; pixel coordinates should be integer!
        virtual v2f      GetAbsoluteSize     () const; // Returns the object's actual size (in pixels) on the screen
        // TODO: Fix math so we can return a Vec2i here; pixel coordinates should be integer!
        virtual v2f      GetAbsolutePosition () const; // Returns the object's actual position (in pixels) on the screen
        float                 GetAbsoluteRotation () const;

        Widget* GetParent       () const;
        UDim2                 GetPosition     () const;
        UDim2                 GetSize         () const;
        float                 GetRotation     () const;
        std::string           GetName         () const;
        UDim                  GetPaddingLeft  () const;
        UDim                  GetPaddingRight () const;
        UDim                  GetPaddingTop   () const;
        UDim                  GetPaddingBottom() const;
        UDim                  GetMarginLeft   () const;
        UDim                  GetMarginRight  () const;
        UDim                  GetMarginTop    () const;
        UDim                  GetMarginBottom () const;
        bool                  GetVisible      () const;
        v2f GetAnchorPoint();
        template<typename T>
        T *FindFirstChildOfType() const;
        bool GetFocused() const;
#pragma endregion
#pragma region Setters
        void SetName            (std::string n);
        void SetParent          (Widget *parent);
        void SetPosition        (UDim2 const &);
        void SetSize            (UDim2 const &);
        void SetRotation        (float rot);
        void SetPaddingLeft     (UDim pl);
        void SetPaddingRight    (UDim pr);
        void SetPaddingTop      (UDim pt);
        void SetPaddingBottom   (UDim pb);
        void SetPadding         (UDim padding);
        void SetMarginLeft      (UDim ml);
        void SetMarginRight     (UDim mr);
        void SetMarginTop       (UDim mt);
        void SetMarginBottom    (UDim mb);
        void SetVisible(bool enabled);
        void SetAnchorPoint(v2f);
        void SetFocused(bool focused);
        void SetNextFocusWidget(Widget* w)
        {
            this->next = w;
        }
        void SetPreviousFocusWidget(Widget* w)
        {
            this->prev = w;
        }
#pragma endregion
#pragma endregion
        virtual void Draw(SDL_Renderer *target);
        virtual void Update(double delta);
        void DrawChildWidgets(SDL_Renderer* r);
        void UpdateChildWidgets(double delta);
        Widget* GetFamilyTreeRoot() const;
#pragma endregion
#pragma region Internals
    protected:
        UDim2 position = {0, 0, 0.0f, 0.0f};
        UDim2 size = {200, 50, 0.0f, 0.0f};
        Widget *parent = nullptr;
        std::vector<Widget *> children;
        float rotation = 0;
        std::string name;
        bool override_child_widget_size;
        bool override_child_widget_position;
        bool selected = false;
        UDim padding_left   = 0_px;
        UDim padding_right  = 0_px;
        UDim padding_top    = 0_px;
        UDim padding_bottom = 0_px;
        UDim margin_left    = 0_px;
        UDim margin_right   = 0_px;
        UDim margin_top     = 0_px;
        UDim margin_bottom  = 0_px;
        v2f anchor_point = {0.f,0.f};
        bool visible = true;
        int descendant_running_tally = 0;

        Widget* next = nullptr;
        Widget* prev = nullptr;
    private:

#pragma endregion
    };


}