#pragma once

#include <glm/glm.hpp>
#include <glm/matrix.hpp>
#include <glm/ext/matrix_common.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <SDL2/SDL.h>
#include <SDL2_gfxPrimitives.h>
#include "JUI/Types/Font.hpp"
#include <JUI/Types/Event.hpp>
#include <JUI/Types/Tween.hpp>

namespace JUI {

    // Aliases for GLM types
    using v2f = glm::vec<2, float>;
    using v2i = glm::vec<2, int>;
    using v2d = glm::vec<2, double>;

    constexpr v2f v2_zero = {0, 0};
    constexpr v2f v2_one = {1, 1};

    using quaternion = glm::quat;

    struct Angle
    {
    public:
        Angle() : radians(0.f) {}
        Angle(float rad) : radians(rad) {}
        Angle(const Angle&) = default;
        float GetDegrees() const { return radians * ( 180.f / std::numbers::pi ); }
        float GetRadians() const { return radians; }

        Angle operator+(const Angle&rhs) const
        { return {this->radians+rhs.radians}; }

        Angle operator-(const Angle&rhs) const
        { return {this->radians-rhs.radians};}

        float radians;
    };

    //

    class Transform2D {
    public:
        Transform2D() : matrix(0) {}
        Transform2D(const v2f& translation, Angle angle, const v2f& scale, const v2f& offset, const v2f& shear)
        {

        }
        Transform2D(const Transform2D&) = default;
        Transform2D(Transform2D&&) = default;
        Transform2D& Reset()
        {
            matrix = {};
            return *this;
        }
        Transform2D& Apply(const Transform2D& rhs);
        Transform2D& Shear(const v2f&);
        Transform2D& ShearX(float x);
        Transform2D& ShearY(float y);
        Transform2D& Skew(const v2f&);
        Transform2D& SkewX(float x);
        Transform2D& SkewY(float y);
        Transform2D& Scale(const v2f& rhs);
        Transform2D& Scale(float scale);
        Transform2D& Scale(float x, float y) { Scale({x, y}); }
        Transform2D& ScaleX(float x);
        Transform2D& ScaleY(float y);
        Transform2D& Translate(const v2f&);
        Transform2D& TranslateX(float x);
        Transform2D& TranslateY(float y);
        Transform2D& Rotate(const Angle&);
        Transform2D& RotateClockwise(const Angle&);
        Transform2D& RotateCCW(const Angle&);
        Transform2D& Pivot(const v2f&, const Angle&);
        Transform2D& Inverse();
        bool IsAffine2DTransform() const;
        v2f TransformPoint(const v2f&);
        v2f InverseTransformPoint(const v2f&);
        glm::mat3x3 GetMatrix() const;

        Transform2D operator*(const Transform2D& rhs);

        Transform2D& operator*=(const Transform2D& rhs);
        bool operator==(const Transform2D& rhs);
        bool operator!=(const Transform2D& rhs);

        Transform2D Lerp(const Transform2D& other, float scalar) const;

    protected:
        glm::mat3x3 matrix;
    private:

    };


    JUI::Angle operator "" _degrees(long double dg);
    JUI::Angle operator "" _radians(long double rd);

    JUI::Angle operator "" _deg(long double dg);
    JUI::Angle operator "" _rad(long double rd);


    enum class ButtonState {Pressed, Released};
    enum class MouseButtonID : uint32_t {NONE = 0, LMB = 1, MMB = 2, RMB = 4, MM4 = 8, MM5 = 16};


    // All this is just for implementing default font
    // Is there a better approach?

    JUI::Font* GetDefaultFont();
    SDL_Texture* GetCheckmark();

    // Basically Wrappers for SDL_Events to provide a more "in-line" API with what we want...
    struct JUIEvent
    {
        void Cancel() {}
    };

    struct MouseMotionEvent : public JUIEvent {
        v2i Delta;
        v2i Position;

        MouseMotionEvent(SDL_MouseMotionEvent e)
        {
            Position = {e.x, e.y};
            Delta = {e.xrel, e.yrel};
        }
    };

    struct MouseButtonEvent : public JUIEvent {
        MouseButtonID Button;
        ButtonState State;
        uint8_t Clicks;
        v2i Position;

        MouseButtonEvent(SDL_MouseButtonEvent e)
        {
            Position = {e.x, e.y};
            Clicks = e.clicks;

            if (e.state == SDL_PRESSED) {
                State = JUI::ButtonState::Pressed;
            } else {
                State == ButtonState::Released;
            }

            using JUI::MouseButtonID;

            if (e.button == SDL_BUTTON_LEFT) {
                Button = MouseButtonID::LMB;
            }
            if (e.button == SDL_BUTTON_RIGHT) {
                Button = MouseButtonID::RMB;
            }
            if (e.button == SDL_BUTTON_MIDDLE) {
                Button = MouseButtonID::MMB;
            }
            if (e.button == SDL_BUTTON_X1) {
                Button = MouseButtonID::MM4;
            }
            if (e.button == SDL_BUTTON_X2) {
                Button = MouseButtonID::MM5;
            }
        }
    };


    static Event<MouseMotionEvent> MouseMoved;
    static Event<MouseButtonEvent> MousePressed;
    static Event<MouseButtonEvent> MouseReleased;

    void Load(SDL_Renderer* renderer);
    void ProcessSDLEvent(SDL_Event event);
    void Update(double delta);
    void Cleanup();

    enum class Keycode
    {
        Escape, F1, F2, F3, F4, F5, F6, F7, F8, F9, F10, F11, F12,
        Tab, Q, W, E, R, T, Y, U, I, O, P, LeftBracket, RightBracket, Backslash,
    };

    struct MouseButtonState {
        bool LMB;
        bool RMB;
        bool MMB;
        bool Mouse4;
        bool Mouse5;
        bool Mouse6;
        bool Mouse7;
    };

    class EventArgs {
    public:
        void Cancel();
        void Consume();
        void Sink() { Consume(); }

    protected:
    private:
    };

    class MouseButtonInputEventArgs {
        MouseButtonID Button;
        ButtonState State;
    };

    struct QuadraticCurve
    {
        v2f p0;
        v2f p1;
        v2f p2;
        v2f p3;
    };

    using curve = QuadraticCurve;

    void RunTween(Tween tween);


    v2i GetWindowSize(SDL_Window* window);

    void SetWindowSize(SDL_Window* window, int x, int y);

    void SetWindowSize(SDL_Window* window, const v2i& size);

    v2i GetMousePosition();

    v2i GetTextureDimensions(SDL_Texture* tex);

    void draw_simple_text(SDL_Renderer* gfx, std::string text, v2i coords);

    void draw_simple_text(SDL_Renderer* gfx, std::string text, int x, int y);


    // Common funcs
    inline float naive_lerp(float a, float b, float t) { return a + t * (b - a); }

    inline SDL_Color lerp_color(SDL_Color a, SDL_Color b, float alpha)
    {
        return
                {
                        static_cast<Uint8>(naive_lerp(static_cast<float>(a.r), static_cast<float>(b.r), alpha)),
                        static_cast<Uint8>(naive_lerp(static_cast<float>(a.g), static_cast<float>(b.g), alpha)),
                        static_cast<Uint8>(naive_lerp(static_cast<float>(a.b), static_cast<float>(b.b), alpha)),
                        static_cast<Uint8>(naive_lerp(static_cast<float>(a.a), static_cast<float>(b.a), alpha))
                };
    }
}