#include <iostream>
#include <chrono>
#include <SDL2/SDL.h>
#include <SDL_image.h>
#include <SDL_video.h>
#include <JUI/jui.h>
#include <Tools/gizmo.hpp>
#include <JUI/Bezier.hpp>
#include <fmt/format.h>
#include <JUI/Widgets/Button.hpp>
#include <JUI/Widgets/Window.h>


// This is a demo of a Minimal SDL2 rendering setup, and how to integrate JUI
// into it as such, as opposed to the abstraction layer of SDLGame.hpp cluttering
// the involved logic
// Basically, simplest demo:

SDL_Window* init_window() {
    SDL_Window* window;
    if (SDL_Init(SDL_INIT_EVERYTHING) < 0)
        std::cerr << "SDL_Error: " << SDL_GetError() << std::endl;

    window = SDL_CreateWindow("SDL Game", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 1000, 600,
                              SDL_WINDOW_SHOWN);

    SDL_SetWindowResizable(window, SDL_TRUE);

    if (window == nullptr) {
        std::cerr << "SDL_Error: " << SDL_GetError() << std::endl;
    }
    return window;
}

SDL_Renderer* init_renderer(SDL_Window* window) {
    SDL_Renderer* renderer;
    renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED);
    if (renderer == nullptr) {
        std::cerr << "SDL_Error: " << "Couldn't init SDL_Renderer." << std::endl;
    }

    SDL_GL_SetSwapInterval(0);
    SDL_RenderSetVSync(renderer, 0);
    return renderer;
}

int main(int argc, char** argv) {
    using namespace std::chrono_literals;
    using precision_clock = std::chrono::high_resolution_clock;
    using precision_timestamp = precision_clock::time_point;
    using seconds_f = std::chrono::duration<float>;
    using seconds_d = std::chrono::duration<double>;

#pragma region SDL Initialization
    SDL_Window   *window   = init_window  ();
    SDL_Renderer *renderer = init_renderer(window);

    int imgFlags = IMG_INIT_PNG;
    if (!(IMG_Init(imgFlags) & imgFlags))
        std::cerr << "SDL_Error: " << "Couldn't init SDL_Image." << std::endl;

#pragma endregion


#pragma region Scene Construction
    JUI::Load(renderer);

    auto *demo_scene = new JUI::Scene(window, renderer);

    auto *test = new JUI::Window(demo_scene);
    test->SetTitle("YUNG MUNEY");

    // TODO: Implement MenuBar widget
    //auto* mnubar = new JUI::MenuBar(test);

    auto *btn = new JUI::Button(test);
    btn->SetBackgroundColor({128, 128, 128, 255});
    btn->SetAnchorPoint({0.f, 0.f});
    btn->SetPosition({25_px, 25_percent});
    btn->SetBorderRadius(12.f);
    btn->SetBorderWidth(1.f);
    btn->SetBorderColor({0, 255, 255, 255});
    btn->SetClipsDescendants(false);
    btn->SetPadding(5_px);
    btn->Center();

    btn->Pressed += [&] (auto _) {
        std::cout << "WILLIAM J TOMASINE II" << std::endl;
    };

    Tween* activeTween = nullptr;


#pragma endregion

    gizmo p0 = gizmo({100, 100});
    gizmo p1 = gizmo({200, 100});
    gizmo p2 = gizmo({100, 200});
    gizmo p3 = gizmo({200, 200});

#pragma region Gameloop
    [[maybe_unused]]
    int frame_count = 0;
    bool request_quit = false;

    precision_timestamp start, end;

    seconds_d total_time {0s};
    seconds_d delta_time {0s};

    while(!request_quit) {
        frame_count++;
        start = precision_clock::now();

        // Handle Events
        SDL_Event event;
        while(SDL_PollEvent(&event)) {
            JUI::ProcessSDLEvent(event);

            if (event.type == SDL_QUIT) {
                request_quit = true;
            }

            if (event.type == SDL_MOUSEMOTION) { }
            if (event.type == SDL_KEYDOWN) { }
            if (event.type == SDL_KEYUP)   { }
        }
#pragma region Update
        JUI::Update(delta_time.count());
        demo_scene->Update(delta_time.count());
        p0.Update(); p1.Update();
        p2.Update(); p3.Update();

#pragma endregion

#pragma region Render

#pragma region Pre Render
        SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
        SDL_RenderClear(renderer);
        auto window_size = GetWindowSize(window);
        SDL_Rect viewport{0, 0, window_size.x, window_size.y};
        SDL_RenderSetViewport(renderer, &viewport);
        SDL_SetRenderDrawColor(renderer, 255, 255, 255, 255);
#pragma endregion

#pragma region Draw Content

        demo_scene->Draw(renderer);

        SDL_Color red {255,   0,   0, 255};
        SDL_Color blu {128, 128, 255, 255};

        JUI::DrawCurve(renderer, red, p0.position, p1.position, p2.position, p3.position, 3);
        JUI::DrawCurve(renderer, blu, p0.position, p1.position, p2.position, p3.position, 30);

        p0.Draw(renderer);
        p1.Draw(renderer);
        p2.Draw(renderer);
        p3.Draw(renderer);

        double dt = delta_time.count();
        //int framerate = SDL_getFramerate(fpsManager);
        auto dt_display = fmt::format("dt: {:.5} ms", dt * 1000);
        stringRGBA(renderer, 0, 0, dt_display.c_str(), 255, 255, 255, 255);

        auto fps_display = fmt::format("fps: {0}", std::floor(1.f / dt));
        stringRGBA(renderer, 0, 10, fps_display.c_str(), 255, 255, 255, 255);

#pragma endregion

#pragma region Post Render
        SDL_RenderPresent(renderer);
#pragma endregion

#pragma endregion

        // TODO: Implement timestep compensation
        //std::this_thread::sleep_for(1ms);
        end = precision_clock::now();
        delta_time = std::chrono::duration_cast<seconds_d>(end - start);
        total_time += delta_time;
    }
    std::cout << "DONE!" << std::endl;
    // Call this upon gameloop exit to dispose of assets
    JUI::Cleanup();

#pragma endregion
}
