#include <cmath>
#include <JUI/Types/Tween.hpp>

namespace JUI {
    void Tween::Update(double delta) {
        progress += delta;

        float total = progress / Runtime;
        Callback(Step(progress));

        if (total > 1) {
            done = true;
            return;
        }
    }

    float Tween::Step(float progress) const {
        float total = progress / Runtime;
        return this->EasingStyle(total, Start, Goal);
    }

    void Tween::Stop() {
        progress = Runtime;
        Step(progress);
        //Done = true;
    }

    float Tween::GetProgress() const { return progress; }



    // References here
    // https://github.com/mobius3/tweeny/blob/master/include/easing.h

    namespace Easing
    {
        float Default(float position, float start, float end) {
            //return (roundf((end - start) * position + start));
            return ((end - start) * position + start);
        }
        float Linear(float position, float start, float end) {
            return roundf((end-start) * position + start);
        }
        float Quadratic::In(float position, float start, float end)
        { return (end - start) * position * position + start; }

        float Quadratic::Out(float position, float start, float end)
        { return ((-(end - start)) * position * (position - 2) + start); }

        float Quadratic::InOut(float position, float start, float end)
        { return ((end - start) / 2) * position * position + start;}

        float Cubic::In(float position, float start, float end) {
            return (end - start) * position * position * position + start;
        }

        float Cubic::Out(float position, float start, float end) {
            return (end - start) * (position * position * position + 1) + start;
        }

        float Cubic::InOut(float position, float start, float end) {
            position *= 2;
            if (position < 1)
                return ((end - start) / 2) * position * position * position + start;
            position -= 2;
            return ((end-start) / 2) * (position * position * position + 2) + start;
        }

        float Quartic::In(float position, float start, float end) {
            return ((end - start) * position * position * position * position + start);
        }

        float Quartic::Out(float position, float start, float end) {
            return (-(end - start) * (position * position * position * position - 1) + start);
        }
        float Quartic::InOut(float position, float start, float end) {
            position *= 2;
            if (position < 1) {
                return ((end - start) / 2) * (position * position * position * position) + start;
            }
            position -= 2;
            return (-(end - start) / 2) * (position * position * position * position - 2) + start;
        }

        float Quintic::In(float p, float start, float end) {
            return ((end - start) * p*p*p*p*p + start);
        }

        float Quintic::Out(float p, float start, float end) {
            return ((end - start) * (p*p*p*p*p + 1) + start);
        }

        float Quintic::InOut(float p, float start, float end) {
            p *= 2;
            if (p < 1)
                return ((end - start) / 2) * (p*p*p*p*p) + start;
            p -= 2;
            return ((end - start) / 2) * (p*p*p*p*p + 2) + start;
        }

        float Sinusoidal::In(float p, float start, float end) {
            return (-(end - start) * cosf(p * M_PI / 2) + (end - start) + start);
        }

        float Sinusoidal::Out(float p, float start, float end) {
            return (end - start) * sinf(p * M_PI / 2) + start;
        }

        float Sinusoidal::InOut(float p, float start, float end) {
            return (-(end - start) / 2) * (cosf(p) * M_PI - 1) + start;
        }

        float Exponential::In(float p, float start, float end) {
            return (end - start) * powf(2, 10 * (p - 1)) + start;
        }

        float Exponential::Out(float p, float start, float end) {
            return ((end - start) * (-powf(2, -10 * p) + 1) + start);
        }

        float Exponential::InOut(float p, float start, float end) {
            p *= 2;
            if (p < 1)
                return ((end - start) / 2) * powf(2, 10* (p-1)) + start;
            --p;
            return ((end - start) / 2) * (-powf(2, -10 * p) + 2) + start;
        }

        float Circular::In(float p, float start, float end) {
            return -(end - start) * (sqrtf(1 - p * p) - 1) + start;
        }

        float Circular::Out(float p, float start, float end) {
            return (end - start) * (sqrtf(1 - p*p)) + start;
        }

        float Circular::InOut(float p, float start, float end) {
            p *= 2;
            if (p < 1)
                return (-(end - start) / 2) * (sqrtf(1 - p*p) - 1) + start;
            p -= 2;
            return ((end - start) / 2) * (sqrtf(1-p*p) + 1) + start;
        }

        float Bounce::In(float p, float start, float end) {
            return (end - start) - Bounce::Out(1 - p, 0, (end-start)) + start;
        }

        float Bounce::Out(float p, float start, float end) {
            float c = end - start;

            if (p < (1 / 2.75f))
                return (c * (7.5625f * p * p) + start);
            else if (p < (2.0f / 2.75f)) {
                float postFix = p -= (1.5f / 2.75f);
                return (c * (7.5625f * (postFix) * p + .75f) + start);
            } else if (p < (2.5f / 2.75f)) {
                float postFix = p -= (2.25f / 2.75f);
                return (c * (7.5625f * (postFix) * p + 0.9375f) + start);
            } else {
                float postFix = p -= (2.625f / 2.75f);
                return (c * (7.5625f * (postFix) * p + 0.984375f) + start);
            }
        }

        float Bounce::InOut(float p, float start, float end) {
            if (p <= 0.00001f) return start;
            if (p >= 0.999f) return end;
            float pp = 0.3f;
            auto a = end - start;
            float s = pp / 4;
            float postFix = a * powf(2, 10 * (p-=1));
            return (-postFix * sinf(p - s) * (2 * M_PI) / pp) + start;
        }

        float Elastic::In(float position, float start, float end) {
            if (position <= 0.00001f) return start;
            if (position >= 0.999f) return end;
            float p = .3f;
            auto a = end - start;
            float s = p / 4;
            return (a * powf(2, -10 * position) * sinf((position - s) * (2 * static_cast<float>(M_PI)) / p) + end);
        }

        float Elastic::Out(float position, float start, float end) {
            if (position <= 0.00001f) return start;
            if (position >= 0.999f) return end;
            position *= 2;
            float p = (.3f * 1.5f);
            auto a = end - start;
            float s = p / 4;
            float postFix;

            if (position < 1) {
                postFix = a * powf(2, 10 * (position -= 1)); // postIncrement is evil
                return (-0.5f * (postFix * sinf((position - s) * (2 * static_cast<float>(M_PI)) / p)) + start);
            }
            postFix = a * powf(2, -10 * (position -= 1)); // postIncrement is evil
            return (postFix * sinf((position - s) * (2 * static_cast<float>(M_PI)) / p) * .5f + end);
        }


        float Elastic::InOut(float position, float start, float end) {
            if (position <= 0.00001f) return start;
            if (position >= 0.999f) return end;
            position *= 2;
            float p = (.3f * 1.5f);
            auto a = end - start;
            float s = p / 4;
            float postFix;

            if (position < 1) {
                postFix = a * powf(2, 10 * (position -= 1)); // postIncrement is evil
                return (-0.5f * (postFix * sinf((position - s) * (2 * static_cast<float>(M_PI)) / p)) + start);
            }
            postFix = a * powf(2, -10 * (position -= 1)); // postIncrement is evil
            return (postFix * sinf((position - s) * (2 * static_cast<float>(M_PI)) / p) * .5f + end);
        }


        float Back::In(float position, float start, float end) {
            float s = 1.70158f;
            float postFix = position;
            return ((end - start) * (postFix) * position * ((s + 1) * position - s) + start);
        }

        float Back::Out(float position, float start, float end) {
            float s = 1.70158f;
            position -= 1;
            return ((end - start) * ((position) * position * ((s + 1) * position + s) + 1) + start);
        }

        float Back::InOut(float position, float start, float end)
        {
            float s = 1.70158f;
            float t = position;
            auto b = start;
            auto c = end - start;
            float d = 1;
            s *= (1.525f);
            if ((t /= d / 2) < 1) return (c / 2 * (t * t * (((s) + 1) * t - s)) + b);
            float postFix = t -= 2;
            return (c / 2 * ((postFix) * t * (((s) + 1) * t + s) + 2) + b);
        }
    }
}

