#include <JUI/Widgets/Rect.hpp>

#include <SDL2_gfxPrimitives.h>
#include <JUI/NumberFormatter.h>

namespace JUI {

    Rect::Rect() : Widget() {
        bg_color = {128, 128, 128, 255};
    }

    Rect::Rect(Widget* parent) : Rect()
    {
        this->SetParent(parent);
    }

    void Rect::Draw(SDL_Renderer* target) {
        if (!visible)
            return;

        v2f abs_pos = GetAbsolutePosition();
        v2f abs_size = GetAbsoluteSize();

        SDL_Rect box{(int)abs_pos.x, (int)abs_pos.y, (int)abs_size.x, (int)abs_size.y};

        auto root_size = GetFamilyTreeRoot()->GetAbsoluteSize();

        // Set Clipping if enabled
        auto* old_clip = new SDL_Rect{0,0,static_cast<int>(root_size.x),static_cast<int>(root_size.y)};

        if (clips_descendants) {
            if (SDL_RenderIsClipEnabled(target))
                SDL_RenderGetClipRect(target, old_clip);

            SDL_RenderSetClipRect(target, &box);
        }

        // Draw Backgrounc
        roundedBoxRGBA(target, box.x, box.y, box.x+box.w, box.y+box.h, border_radius, bg_color.r, bg_color.g, bg_color.b, bg_color.a);


        // Draw Outline

        // TODO: Implement custom outline (again) using thickLine and bezierCurve for border?
        // SDL_gfx doesn't let you set the line width for roundedRectangle...
        if (border_radius >= 1) {
            // Draw Background
            for (int i = 0; i < border_width; i++) {
                roundedRectangleRGBA(target, (box.x-i), (box.y-i), box.x + box.w + (i*2), box.y + box.h + (i*2), border_radius, border_color.r,
                                     border_color.g, border_color.b, border_color.a);

            }
        } else {
            // TODO: Implement non-fucked-up border thickness rendering.

            for (int i = 0; i < border_width; i++) {
                rectangleRGBA(target, (box.x), (box.y), (box.x + box.w - (i * 2)) + 1, (box.y + box.h - (i * 2)) + 1,
                              border_color.r,
                              border_color.g, border_color.b, border_color.a);
            }
        }

        Widget::Draw(target);

        // Return clip to previous state
        if (clips_descendants && SDL_RenderIsClipEnabled(target)==SDL_TRUE)
            SDL_RenderSetClipRect(target, old_clip);
    }

    void Rect::Update(double delta)
    {
        Widget::Update(delta);
        int x;
        int y;
        Uint32 mouse_btn_bitmask = SDL_GetMouseState(&x, &y);
        // TODO: Implement easy conversion between JUI::Vector2<T> and SDL_Point
        const SDL_Point mouse_coords{x, y};
        auto abs_size = GetAbsoluteSize();
        auto abs_pos = GetAbsolutePosition();

        MouseButtonState state {};

        state.LMB = mouse_btn_bitmask & SDL_BUTTON(1);
        state.MMB = mouse_btn_bitmask & SDL_BUTTON(2);
        state.RMB = mouse_btn_bitmask & SDL_BUTTON(3);

        // TODO: Pull this out into some kind of InputService singleton
        if (!prev_state.LMB && state.LMB && mouse_hover) {
            was_clicked = true;
            MousePressed.Invoke(MouseButtonID::LMB);
        }

        if (prev_state.LMB && !state.LMB && was_clicked) {
            MouseReleased.Invoke(MouseButtonID::LMB);
            was_clicked = false;
        }

        if (!prev_state.RMB && state.RMB && mouse_hover) {
            was_clicked = true;
            MousePressed.Invoke(MouseButtonID::RMB);
        }

        if (prev_state.RMB && !state.RMB && was_clicked) {
            MouseReleased.Invoke(MouseButtonID::RMB);
            was_clicked = false;
        }

        if (!prev_state.MMB && state.MMB && mouse_hover) {
            was_clicked = true;
            MousePressed.Invoke(MouseButtonID::MMB);
        }

        if (prev_state.MMB && !state.MMB && was_clicked) {
            MouseReleased.Invoke(MouseButtonID::MMB);
            was_clicked = false;
        }


        prev_state = state;


        SDL_Rect bounds{
                static_cast<int>(abs_pos.x),
                static_cast<int>(abs_pos.y),
                static_cast<int>(abs_size.x),
                static_cast<int>(abs_size.y)};

        if (IsMouseInside()) { //(SDL_PointInRect(&mouse_coords, &bounds)) {
            if (!mouse_hover) {
                const v2f fcoords = {mouse_coords.x, mouse_coords.y};
                MouseEnter.Invoke(fcoords);
            }
            mouse_hover = true;
        } else {
            if (mouse_hover) {
                const v2f fcoords = {mouse_coords.x, mouse_coords.y};
                MouseExit.Invoke(fcoords);
            }
            mouse_hover = false;
        }
    }

    SDL_Color Rect::GetBackgroundColor() const
    { return bg_color;}

    void Rect::SetBackgroundColor(SDL_Color c)
    { bg_color = c;}

    SDL_Color Rect::GetBorderColor() const
    { return border_color;}

    void Rect::SetBorderColor(SDL_Color c)
    { border_color = c;}

    float Rect::GetBorderWidth() const
    { return border_width; }



    void Rect::SetBorderWidth(float pixels)
    { border_width = pixels; }

    float Rect::GetBorderRadius() const
    { return border_radius;}

    void Rect::SetBorderRadius(float radius)
    { border_radius = radius; }

    bool Rect::GetClipsDescendants() const
    { return clips_descendants; }

    void Rect::SetClipsDescendants(bool ye)
    { clips_descendants = ye;}

    // TODO: FIX
    bool Rect::IsMouseInside() const {
        int x, y;
        SDL_GetMouseState(&x, &y);

        auto pos = GetAbsolutePosition();
        auto size = GetAbsoluteSize();


        if (x > pos.x && y > pos.y && x < pos.x + size.x && y < pos.y + size.y) {
            return true;
        }
        return false;
    }





}